/*
 * stat: show stat fields for files.
 * 92/03/31 john h. dubois iii (john@armory.com)
 * 92/05/09 added m and c options; expanded help
 * 92/09/24 exit(0) at end of main(); added r option.
 * 93/10/01 Added t option, A, M, and C field specs, and useful exit status.
 * 94/02/21 Added d and o options.
 * 94/04/18 Added b, N, J field specs.  Substantially reorganized.
 * 94/10/08 Added l option, and LOB field specs.  Fixed bug that caused field
 *          sep to be printed even if first field was skipped.
 *          Do not print S field for block devices.
 *          Print newline before perrors that might not be at start of line.
 * 94/10/13 Let int char values given with -c.
 * 94/12/18 Added L option.
 * 95/07/29 Print name before all error messages.
 * 95/08/16 Added s option and t field field.
 * 95/09/22 Made -a include OB specs if -m also given.
 *          Allow symbolic mode chars to be given with -m.
 *          Added DMT options.
 * 95/10/02 Cleaned up declarations.  Don't do statfs for block size if we
 *          have st_stat32.  Do at most one statfs per file.
 * 96/01/08 Flush stdout before writing to stderr for errors that are liable
 *          to occur in the middle of normal output, to keep output separate
 *          when e.g.  2>&1 | ...  is done.
 * 96/06/06 Allow null as tag character.
 */

/*
*   To add a field printing option to -f, must:
*   Add a field printing function prototype at 'Field printing prototypes'
*   Add a field printing function at 'Field print functions'
*   Add an option letter to function mapping line to InitFuncs()
*   If field s/b printed by default, add option letter to Fields decl in main()
*   Add option letter to AllFields[] decl in main()
*   Add field description to Help[] in main()
*/
/*
    todo: add field specs for the info from a statfs() on the st_dev the file
    resides on, and the info from a pathconf() on the file.
    Let the output be formatted using a string of %c field specifiers.
*/
char *Comment = "@(#) stat 1.5 96/06/06";

#ifdef _M_SYSV
#define M_SYSV
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/statfs.h>
#include <stdio.h>
#include <dirent.h>
#include <time.h>
#include <stdlib.h>
#include <limits.h>
#include <unistd.h>
#include <sys/errno.h>
#include <pwd.h>
#include <grp.h>
#include <dirent.h>
#ifdef M_SYSV
#include <string.h>
#else
#include <strings.h>
#endif

/* the best preprocessor test I could find in sys/stat.h */
#if (_SCO_STAT_VER == 51)
#define HAVESTAT32
/* This is not strictly neccessary since struct stat is made the same as
   struct st_stat32, but it ensures that the compile will fail if for some
   reason struct st_stat32 is not declared */
typedef struct st_stat32 stat_t;
#else
typedef struct stat stat_t;
#endif

#define const	/* header files don't match man page */

/* Field printing prototypes */
char *PrintTime(time_t Time);
char *P_linkname(stat_t *buf,const char **OutBuf);
char *P_ino(stat_t *buf,const char **OutBuf);
char *P_dev(stat_t *buf,const char **OutBuf);
char *P_nlink(stat_t *buf,const char **OutBuf);
char *P_linkValue(stat_t *buf,const char **OutBuf);
char *P_uid(stat_t *buf,const char **OutBuf);
char *P_gid(stat_t *buf,const char **OutBuf);
char *P_size(stat_t *buf,const char **OutBuf);
char *P_Major(stat_t *buf,const char **OutBuf);
char *P_Minor(stat_t *buf,const char **OutBuf);
char *P_atime(stat_t *buf,const char **OutBuf);
char *P_mtime(stat_t *buf,const char **OutBuf);
char *P_ctime(stat_t *buf,const char **OutBuf);
char *P_Atime(stat_t *buf,const char **OutBuf);
char *P_Mtime(stat_t *buf,const char **OutBuf);
char *P_Ctime(stat_t *buf,const char **OutBuf);
char *P_blocks(stat_t *buf,const char **OutBuf);
char *P_ngid(stat_t *buf,const char **OutBuf);
char *P_nuid(stat_t *buf,const char **OutBuf);
char *P_mode(stat_t *buf,const char **OutBuf);
char *P_symmode(stat_t *buf,const char **OutBuf);
char *P_Xmode(stat_t *buf,const char **OutBuf);
char *P_Xsymmode(stat_t *buf,const char **OutBuf);
char *P_tag(stat_t *buf,const char **OutBuf);
int InitFuncs(char *(*Funcs[])(), char *S, int gotTag);
int StatFile(char *File, int fd, char *(*Funcs[])(), char FieldChar,
int NameLen, int Desc,int ReadDirs);
int StatAllOpen(char *(*Funcs[])(),char FieldChar,int NameLen);
struct statfs *DoStatfs();
void symmodeCommon(stat_t *StatB, const char **OutB, mode_t PrintModes);
mode_t GetMode(char *ModeString, char OptChar);
int FileType(char c, mode_t *Mode, dev_t *RDev);

#define MAXTIME 1024
#define TIME_FMT "%y/%m/%d %T"
#define BUFFER_SIZE (MAXNAMLEN+1)
#define MAXTAG 1024

mode_t ModeMask = (mode_t) ~(0L);	/* Initially include all mode bits */
/* Mask to AND mode with to determine whether to print file info */
mode_t CheckModeMask = 0;
int ModeZeros = 1;  /* Width of field to print for o field spec (P_mode()) */
int ModeXZeros = 1;  /* Width of field to print for O field spec (P_Xmode()) */
int AnyBits = 0;	/* Print file info if any -M masked bits are set */
int AllBits = 0;	/* Print file info if all -M masked bits are set */
int CheckType = 0;	/* Print file info only if correct type */
mode_t CheckTypeMode;	/* Print file info only if type matches this */
dev_t CheckTypeRDev;	/* rdev part of file type to match */
int (*StatFunc)() = lstat;
char *lrindex(char *s, char c, unsigned len);

/* If FileName is non-null, a file is being stat()ed by name.
   If it is null, a file is being fstat()ed by file descriptor. */
char *FileName;		/* For use by P_linkname & P_blocks */
int FileNum;		/* For use by P_linkname & P_blocks */
/* Initialize Tag to a null string so that if -s is given along with filenames
   on the command line, the named files will get a null tag. */
char *Tag = "";
int DidStatfs;

char *TimeFormat = TIME_FMT;
extern int errno;

/* Field printing functions */
char *P_tag(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    *OutBuf = Tag;
    return "Tag";
}

char *P_linkname(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    if (FileName)
	*OutBuf = FileName;
    else
        sprintf(*OutBuf,"%d",FileNum);
    return FileName ? "Linkname" : "Handle";
}

char *P_mode(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    sprintf(*OutBuf,
    "%0*o",ModeZeros,(unsigned) ((StatBuf->st_mode) & ModeMask));
    return "Mode";
}

char *P_Xmode(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    mode_t Mode = StatBuf->st_mode;   /* For testing bits in */
    mode_t Mask = ModeMask;	/* Duplicate of mask so it can be shifted */
    mode_t Result = 0;		/* 'squeezed' result */
    mode_t ResultBit = 1;	/* Tracks the result bit to be or'ed into */

    /* Do until all mode bits masked in have been processed */
    while (Mask) {
	if (Mask & 1) {	/* If this mode bit is masked in */
	    if (Mode & 1)
		Result |= ResultBit;
	    ResultBit <<= 1;
	}
	Mask >>= 1;
	Mode >>= 1;
    }
    sprintf(*OutBuf,"%0*o",ModeXZeros,Result);
    return "XMode";
}

/* 
   cc whines at me if I don't declare it this way.  All of the functions
   should really be declared this way to be compatible with their prototypes.
   Some day.
*/
/*
    PrintModes contains a mask describing which mode bits should be included
    in the output.
*/
void symmodeCommon(stat_t *StatBuf, const char **OutBuf,
mode_t PrintModes)
{
    char *p = *OutBuf;
    mode_t Mode = StatBuf->st_mode & ModeMask;
    int i;

    /* If all file type bits are included */
    if ((PrintModes & S_IFMT) == S_IFMT) {
	switch (Mode & S_IFMT) {
	case S_IFREG:
	    *p = 'r';
	    break;
	case S_IFBLK:
	    *p = 'b';
	    break;
	case S_IFDIR:
	    *p = 'd';
	    break;
	case S_IFCHR:
	    *p = 'c';
	    break;
	case S_IFIFO:
	    *p = 'p';
	    break;
	case S_IFNAM:
	    switch (StatBuf->st_rdev) {
	    case S_INSEM:
		*p = 's';
		break;
	    case S_INSHD:
		*p = 'm';
		break;
	    default:
		*p = 'n';
	    }
	    break;
	case S_IFLNK:
	    *p = 'l';
	    break;
	default:
	    *p = '?';
	}
	p++;
    }
    if (PrintModes & S_ISUID)
	*(p++) = (Mode & S_ISUID) ? 'u' : '-';
    if (PrintModes & S_ISGID)
	*(p++) = (Mode & S_ISGID) ? 'g' : '-';
    if (PrintModes & S_ISVTX)
	*(p++) = (Mode & S_ISVTX) ? 't' : '-';
    for (i = 1; i <= 3; i++) {
	if (PrintModes & S_IREAD)
	    *(p++) = (Mode & S_IREAD) ? 'r' : '-';
	if (PrintModes & S_IWRITE)
	    *(p++) = (Mode & S_IWRITE) ? 'w' : '-';
	if (PrintModes & S_IEXEC)
	    *(p++) = (Mode & S_IEXEC) ? 'x' : '-';
	Mode <<= 3;
	PrintModes <<= 3;
    }
    *p = '\0';
}

long GetBlockSize(stat_t *StatBuf)
{
    struct statfs *FS;

#ifdef HAVESTAT32
    if (StatBuf->st_blksize > 0)
	return StatBuf->st_blksize;
#endif

    /* Do statfs() on the filesystem that the file resides on,
       to get its block size. */
    if (!(FS = DoStatfs()))
	return 0;
    return FS->f_bsize;
}

char *P_symmode(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    symmodeCommon(StatBuf,OutBuf,(mode_t) (~0));
    return "Modes";
}

char *P_Xsymmode(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    symmodeCommon(StatBuf,OutBuf,ModeMask);
    return "XModes";
}

char *P_perm(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    sprintf(*OutBuf,"%04o",(unsigned) ((StatBuf->st_mode) & 07777));
    return "Perm";
}

char *P_ino(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    sprintf(*OutBuf,"%u",(unsigned) (StatBuf->st_ino));
    return "Inode";
}

char *P_dev(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    sprintf(*OutBuf,"%u",(unsigned) (StatBuf->st_dev));
    return "Device";
}

char *P_linkValue(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    int FNameLen;

    if (FileName && (FNameLen = readlink(FileName,*OutBuf,BUFFER_SIZE-1)) > 0){
	(*OutBuf)[FNameLen] = '\0';
	return "LinkVal";
    }
	return NULL;
}

char *P_nlink(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    sprintf(*OutBuf,"%u",(unsigned) (StatBuf->st_nlink));
    return "Links";
}

char *P_nuid(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    /* cache last entry params.  This asumes that only this function will
       be calling a getpw routine. */
    static struct passwd *PWent = NULL;
    static int LastUID = -1;
    int UID;

    if ((UID = StatBuf->st_uid) != LastUID)
	PWent = getpwuid(LastUID = UID);
    if (PWent)
	*OutBuf = PWent->pw_name;
    else
	sprintf(*OutBuf,"%u",(unsigned) UID);
    return "User";
}

char *P_uid(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    sprintf(*OutBuf,"%u",(unsigned) (StatBuf->st_uid));
    return "UID";
}

char *P_ngid(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    /* cache last entry params.  This asumes that only this function will
       be calling a getgr routine. */
    static struct group *GRent = NULL;
    static int LastGID = -1;
    int GID;

    if ((GID = StatBuf->st_gid) != LastGID)
	GRent = getgrgid(LastGID = GID);
    if (GRent)
	*OutBuf = GRent->gr_name;
    else
	sprintf(*OutBuf,"%u",(unsigned) GID);
    return "Group";
}

char *P_gid(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    sprintf(*OutBuf,"%u",(unsigned) (StatBuf->st_gid));
    return "GID";
}

char *P_size(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    switch (StatBuf->st_mode & S_IFMT) {
    case S_IFBLK:
    case S_IFCHR:
    case S_IFNAM:
        return NULL;
    }
    sprintf(*OutBuf,"%u",(unsigned) (StatBuf->st_size));
    return "Size";
}

char *P_blocks(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    int BlockSize;

    switch (StatBuf->st_mode & S_IFMT) {
    case S_IFCHR:
    case S_IFBLK:
    case S_IFNAM:
        return NULL;
    }

    if (!(BlockSize = GetBlockSize(StatBuf)))
	return NULL;
    sprintf(*OutBuf,"%u",
    (unsigned) ((StatBuf->st_size + BlockSize - 1) / BlockSize));
    return "Blocks";
}

char *P_Major(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    switch (StatBuf->st_mode & S_IFMT) {
    case S_IFBLK:
    case S_IFCHR:
    case S_IFNAM:
	sprintf(*OutBuf,"%u",(unsigned) (StatBuf->st_rdev) >> 8);
	return "Major";
    }
    return NULL;
}

char *P_Minor(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    switch (StatBuf->st_mode & S_IFMT) {
    case S_IFBLK:
    case S_IFCHR:
    case S_IFNAM:
	sprintf(*OutBuf,"%u",(unsigned) (StatBuf->st_rdev) & 0xFF);
	return "Minor";
    }
    return NULL;
}

size_t PrintFTime(time,Buf)
time_t time;
char *Buf;
{
    return strftime(Buf,(size_t) MAXTIME,TimeFormat,localtime(&time));
}

char *P_Atime(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    PrintFTime((time_t) (StatBuf->st_atime),*OutBuf);
    return "Accessed";
}

char *P_Mtime(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    PrintFTime((time_t) (StatBuf->st_mtime),*OutBuf);
    return "Modified";
}

char *P_Ctime(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    PrintFTime((time_t) (StatBuf->st_ctime),*OutBuf);
    return "Created";
}

#define PrintTime(time,buf) sprintf(buf,"%u",time)

char *P_atime(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    PrintTime((time_t) (StatBuf->st_atime),*OutBuf);
    return "Accessed";
}

char *P_mtime(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    PrintTime((time_t) (StatBuf->st_mtime),*OutBuf);
    return "Modified";
}

char *P_ctime(StatBuf,OutBuf)
stat_t *StatBuf;
const char **OutBuf;
{
    PrintTime((time_t) (StatBuf->st_ctime),*OutBuf);
    return "Created";
}

/*
*   OBt must remain at the end.  They are included in AllFields because it
*   is used as a list of all possible field specs, and to reserve space for 
*   themselves in AllFields[] in case -a is given with -m (we then print OB
*   specs) or -s (we then print t spec).
*   See InitFuncs() for an ordered list of all used letters.
*/
char AllFields[] = "nLpidlobuUgGsSNJamcAMCOBt";

/* Don't count null term char */
#define MAXFUNCS (sizeof(AllFields)-1)

/*
Possible field specs to add:
f: all of statfs
T: statfs.f_fstype;	 File system type
z: statfs.f_bsize;         Block size
Z: statfs.f_frsize;        Fragment size
k: statfs.f_blocks;        Total number of blocks
K: statfs.f_bfree;         Count of free blocks
I: statfs.f_files;         Total number of file nodes
D: statfs.f_ffree;         Count of free file nodes
v: statfs.f_fname[6];      Volume name
P: statfs.f_fpack[6];      Pack name
/: All of pathconf
x: LINK_MAX
X: MAX_CANON
j: MAX_INPUT
e: NAME_MAX
h: PATH_MAX
F: PIPE_BUF
r: CHOWN_RESTRICTED
T: NO_TRUNC
V: VDISABLE
Remaining easily used chars (no % because we will implement %c specs sometime):
EHqQRwWyY@-_=+:,.
*/
int InitFuncs(char *(*Funcs[])(), char *S, int gotTag)
{
    int Pos;
    for (Pos = 0; *S && (Pos < MAXFUNCS); S++) {
	switch (*S) {
	case 'a': Funcs[Pos++] = P_atime; break;
	case 'A': Funcs[Pos++] = P_Atime; break;
	case 'b': Funcs[Pos++] = P_symmode; break;
	case 'B': Funcs[Pos++] = P_Xsymmode; break;
	case 'c': Funcs[Pos++] = P_ctime; break;
	case 'C': Funcs[Pos++] = P_Ctime; break;
	case 'd': Funcs[Pos++] = P_dev; break;
	case 'g': Funcs[Pos++] = P_gid; break;
	case 'G': Funcs[Pos++] = P_ngid; break;
	case 'i': Funcs[Pos++] = P_ino; break;
	case 'J': Funcs[Pos++] = P_Major; break;
	case 'l': Funcs[Pos++] = P_nlink; break;
	case 'L': Funcs[Pos++] = P_linkValue; break;
	case 'm': Funcs[Pos++] = P_mtime; break;
	case 'M': Funcs[Pos++] = P_Mtime; break;
	case 'n': Funcs[Pos++] = P_linkname; break;
	case 'N': Funcs[Pos++] = P_Minor; break;
	case 'o': Funcs[Pos++] = P_mode; break;
	case 'O': Funcs[Pos++] = P_Xmode; break;
	case 'p': Funcs[Pos++] = P_perm; break;
	case 's': Funcs[Pos++] = P_size; break;
	case 'S': Funcs[Pos++] = P_blocks; break;
	case 't':
	    if (!gotTag) {
		fprintf(stderr,
		"stat: Must give -s<char> if 't' field spec is used.\n",*S);
		return(1);
	    }
	    Funcs[Pos++] = P_tag;
	    break;
	case 'u': Funcs[Pos++] = P_uid; break;
	case 'U': Funcs[Pos++] = P_nuid; break;
	default:
	    fprintf(stderr,"stat: %c: invalid field spec.\n",*S);
	    return(1);
	}
    }
    if (*S) {
	fprintf(stderr,"stat: Too many field specs.\n");
	return(1);
    }
    Funcs[Pos] = NULL;
    return 0;
}

/*
Desc: Take all filenames that consist entirely of digits to be file
      descriptors.
fd: File descriptor to stat.
*/
int StatFile(char *File,int fd,char *(*Funcs[])(),char FieldChar,int NameLen,
int Desc, int ReadDirs)
{
    char *(**Func)();
    stat_t buf;
    int Err = 0;
    int DoFd;
    int MustExist = 1;
    char *ret;
    char Buffer[BUFFER_SIZE];
    char *OutBuf;
    int FieldPrinted;
    DIR *Dir;
    struct dirent *Entry;
    mode_t Type;

    FileName = File;
    if (File) {
	/* If allowing file descriptors, and filename isn't null, and consists
	   entirely of digits, do fstat; else stat */
	if (DoFd = Desc && *File && !File[strspn(File,"0123456789")])
	    fd = atoi(File);
    }
    else {
	DoFd = 1;
	MustExist = 0;
	FileNum = fd;
    }

    /* The explicit use of struct st_stat32 conflicts with the declaration
       of fstat(), so cast */
    if ((DoFd ? fstat(fd,(struct stat *)&buf) : (*StatFunc)(File,&buf)) == -1) {
	if (!MustExist && errno == EBADF)
	    return 0;
	fflush(stdout);
	fputs("stat: ",stderr);
	perror(File);
	Err = 1;
    }
    else {
	if (ReadDirs && !DoFd && (buf.st_mode & S_IFMT) == S_IFDIR) {
	    if (!(Dir = opendir(File))) {
		fflush(stdout);
		fputs("stat: could not open directory ",stderr);
		perror(File);
		Err = 1;
	    }
	    else
		while (Entry = readdir(Dir)) {
		    File = Entry->d_name;
		    if (!strcmp(File,".") || !strcmp(File,".."))
			continue;
		    Err |= StatFile(File,0,Funcs,FieldChar,NameLen,0,0);
		}
	    closedir(Dir);
	    return Err;
	}
	if (CheckModeMask) {
	    if (!(AnyBits && (buf.st_mode & CheckModeMask) || 
	    AllBits && ((buf.st_mode & CheckModeMask) == CheckModeMask) || 
	    buf.st_mode == CheckModeMask))
		return 0;
	}
	if (CheckType) {
	    Type = buf.st_mode & S_IFMT;

	    if (Type != CheckTypeMode ||
	    Type == S_IFNAM && buf.st_rdev != CheckTypeRDev)
		return 0;
	}
	FieldPrinted = 0;
	DidStatfs = 0;
	for (Func = Funcs; *Func; Func++) {
	    OutBuf = Buffer;
	    if ( (ret = (*Func)(&buf,&OutBuf)) ) {
		if (FieldPrinted)	/* If not the first field */
		    putchar(FieldChar);
		else
		    FieldPrinted = 1;
		if (NameLen)	/* If field name is to be printed */
		    printf("%-*s  ",NameLen,ret);
		fputs(OutBuf,stdout);
	    }
	}
	putchar('\n');
    }
    return Err;
}

/*
*   If S is 0-length, returns 0.
*   If S is 1 char long, returns that char as an int.
*   If S is longer than 1 char, attempts to convert it as an integer value
*   according to the rules of strtol(,,0), and returns that value.
*   On success, the return value will be an integer in the range 0..255.
*   On failure, -1 is returned.
*   Failure will occur if S is a null pointer, if integer conversion results in
*   a value outside the range 0..255, if strtol() is unable to produce an
*   integer from a string, or if there are characters left over after strtol()
*   does its conversion.
*/
int StrToChar(char *S)
{
    long Ret;
    char *Ptr;	/* For checking if all chars were used in conversion */

    if (!S)
	return -1;
    if (!*S)
	return 0;
    if (!S[1])
	return (unsigned char) *S;
    Ret = strtol(S,&Ptr,0);
    if (!(0 <= Ret && Ret <= 255) || *Ptr)
	return -1;
    return (unsigned char) Ret;
}

int FileType(char c, mode_t *Mode, dev_t *RDev)
{
    *RDev = 0;
    switch (c) {
    case 'r': *Mode = S_IFREG; break;
    case 'b': *Mode = S_IFBLK; break;
    case 'd': *Mode = S_IFDIR; break;
    case 'c': *Mode = S_IFCHR; break;
    case 'p': *Mode = S_IFIFO; break;
    default: return 0;
    }
    return 1;
}

mode_t GetMode(char *ModeString, char OptChar)
{
    mode_t ModeMask = 0;
    char *ret;
    int Error = 0;
    dev_t RDev;
    mode_t Mode;

    for (; *ModeString && !('0' <= *ModeString && *ModeString <= '7');
    ModeString++)
	switch (*ModeString) {
	case 'g': ModeMask |= S_ISGID; break;
	case 'u': ModeMask |= S_ISUID; break;
	case 't': ModeMask |= S_ISVTX; break;
	case 'T': ModeMask |= S_IFMT; break;
	case 'R': ModeMask |= S_IRUSR|S_IRGRP|S_IROTH; break;
	case 'W': ModeMask |= S_IWUSR|S_IWGRP|S_IWOTH; break;
	case 'X': ModeMask |= S_IXUSR|S_IXGRP|S_IXOTH; break;
	case 'U': ModeMask |= S_IRWXU; break;
	case 'G': ModeMask |= S_IRWXG; break;
	case 'O': ModeMask |= S_IRWXO; break;
	case 'n': ModeMask |= S_IFNAM; break;
	case 'b': case 'c': case 'd': case 'p': case 'r':
	    if (FileType(*ModeString,&Mode,&RDev)) {
		ModeMask |= Mode;
		break;
	    }
	default: 
	    fprintf(stderr,
	    "stat: Unknown character passed with -%c: %c\n",
	    OptChar,*ModeString);
	    Error = 1;
	}
    /* Do this only if there are chars left, because strtol will
       return an error for empty strings */
    if (*ModeString) {
	ModeMask |= strtol(ModeString,&ret,8);
	if (ret == ModeString) {
	    fprintf(stderr,
	    "stat: Bad octal value passed with -%c.\n",OptChar);
	    Error = 1;
	}
    }
    if (!ModeMask) {
	fprintf(stderr, "stat: empty mask passed with -%c\n", OptChar);
	Error = 1;
    }
    return Error ? 0 : ModeMask;
}

main(argc,argv)
int argc;
char *argv[];
{
    int i;
    char *(*Funcs[MAXFUNCS + 1])();
    extern char *optarg;
    extern int optind;
    int c;
    /* Fields to print; initialize to default */
    char *Fields = "noidlugsJNamc";
    int Error = 0;
    int Desc = 0;
    int AllOpen = 0;
    int ReadInput = 0;
    int NameLen = 8;	/* Longest field name is 8 chars */
    /* Let us read a maximum-length path + separator-char + maximum-length tag
       and still have room for the null terminator. */
    char FileName[MAXNAMLEN + MAXTAG + 2];
    int Check;
    int aOpt = 0;
    int ReadDirs = 0;	/* -D flag given; stat all files in dirs */
    char tagChar;
    int gotTag = 0;
    char *TagStart;
    int Length;
    char *Usage = "Usage:\n" \
/* Used: acdDfhlLmMnorst */
"stat [-adhlnor] [-[mM]<mask>] [-T<type>] [-s<char>] [-c<char>] [-t<fmt>]\n"\
"     [-[fL][%s]] [file ...]\n";
    char Help1[] =
"No field name or value is printed for fields that have no meaning for the\n"\
"file being described (for example, link contents for a file that is not a\n"\
"symbolic link).\n"\
"Options:\n"\
"n: Don't print stat field names.\n" \
"h: Print this help.\n" \
"r: stat filenames read from stdin, one per line, after any on command line.\n"\
"   Empty lines are ignored.\n"\
"s: Like -r, except that an additional 'tag' string may be given after the\n"\
"   filename.  The tag is separated from the filename to be stat'ed by the\n"\
"   last instance of <char> that occurs on the input line.  Give a separate\n"\
"   null argument (i.e. -s \"\") to set <char> to be a null character (this\n"\
"   is useful because a null cannot appear in a path).  The tag may be\n"\
"   printed with the 't' field specifier.  The tag should be no more than\n"\
"   %d characters.\n"\
"D: For any files that are directories, report on all of the files in the\n"\
"   directory except '.' and '..' rather than on the directory itself.\n"\
"   This works with filenames only, not file descriptors.  Give -l if\n"\
"   directories pointed to by symbolic links should be processed.\n"\
"l: Dereference symbolic links.  Normally, symbolic links are printed as\n"\
"   such.  If -l is given, information about the file the symbolic link\n"\
"   points to is printed.\n"\
"m: Logically AND file mode with <mask> before printing it.\n"\
"   <mask> consists of symbolic mode characters and/or an octal number.\n"\
"   If both are given, the symbolic characters must come first.\n"\
"   The symbolic characters and the mode bits they stand for are:\n"\
"   u: setuid.  g: setgid.  t: sticky.  T: all of the file type bits.\n"\
"   R, W, X: all of the readability, writability, and executability bits.\n"\
"   U, G, O: all of the user-, group-, and other-accessibility bits.\n"\
"   b, c, d, n, p, r: set the file type bits for block, character,\n"\
"   directory, any name type, pipe, and regular files.\n"\
"M: Print file information only if <mask> (specified as for -m) exactly\n"\
"   matches the file mode.  If <mask> is prefixed with a '+' or '-', file\n"\
"   information is printed if any or all of the bits specified by <mask>\n"\
"   are set respectively.\n"\
"T: Print file information only if the file type is <type>.  Possible file\n"\
"   file types are r (regular), d (directory), c (character), b (block),\n"\
"   p (pipe), s (semaphore), and m (memory).\n"
"   they do for the 'b' field speficiation.  ? refers to any type that is\n"\
"   not one of the others.\n"\
"c: Use <char> instead of newline to separate each field from the next.  If\n"\
"   <char> is more than one character long, it is interpreted as an integer\n"\
"   representation of the character, decimal by default.  Octal and hex\n"\
"   values can be given by preceding the value with 0 or 0x respectively.\n"\
"   To separate the fields with null chars, use  -c ''  or  -c 00\n"\
"d: All filenames that consist of digits are taken to be file descriptors\n"\
"   to be fstat()ed.  The file handle is printed instead of the link name.\n"\
"o: fstat() all open file descriptors after other filenames are processed.\n"\
"t: Use <format> as an strftime(S) format string to print file times for\n"\
"   A, M, and C field specifications.  The default is \"%s\".\n"\
"a: Print all fields in all formats.  Equivalent to -f%s.\n"\
"   The OB fields are printed only if -m is given, and the t field is\n"\
"   printed only if -s is given.\n";
    char Help2[] =
"f: Select fields to print.  The default is nMidlugsJNamc.\n"\
"   Fields are printed in the order given.  The fields are:\n"\
"   n: name of the link used to access the file (not part of stat struct).\n"\
"   o: file mode, in octal.  If a mask is given, at least as many\n"\
"      characters as are given in the mask will be printed, with leading\n"\
"      zeros if neccessary.\n" \
"   O: like 'o', except that if a mask is given, any bit excluded is not\n"\
"      printed (the masked-in bits are concatenated together before\n"\
"      conversion to octal).\n"\
"   p: file permissions, in octal (equivalent to -fo -m7777).\n"\
"   b: file mode, symbolically.  The first character indicates the type of\n"\
"      the file (Regular, Directory, Character, Block, Pipe, Semaphore,\n"\
"      shared Memory, other Name file, or ?unknown); the next three tell\n"\
"      whether the setUid, setGid, and sTicky bits are set; and the last\n"\
"      nine tell whether Read, Write, & eXecute are set for user, group,\n"\
"      and other.  If a mask is given and some of the file type bits are\n"\
"      excluded, a bogus type may be printed.\n"\
"   B: like 'b', except that if a mask is given, the character for any bit\n"\
"      excluded is not printed.  If any of the file type bits is excluded,\n"\
"      the file type is not printed.\n"\
"   i: inode number.\n" \
"   d: ID of device containing directory entry.\n"\
"   l: number of links to file.\n" \
"   L: symbolic link contents (file name that symlink points to).\n"\
"   u, g: file's user and group ownership, as integers.\n" \
"   U, G: file's user and group as names if they have entries in /etc/passwd\n"\
"      and /etc/group respectively; else as integers.\n"\
"   s: file size, in bytes (not printed for block, char, or name files).\n"\
"   J, N: Major & minor number (printed only for block, char & name files).\n"\
"   S: file size, in blocks of size used by the filesystem the file resides on.\n"\
"   a, m, c: time of last access, modification, and creation, in seconds\n"\
"      since 00:00:00 GMT 1 Jan 1970.\n"\
"   A, M, C: time of last access, modification, and creation, in local time.\n"\
"L: Like -f, but also sets other modes commonly used for printing status of\n"\
"   files one per line.  Sets -n (stat field names not printed) and -c' '\n"\
"   (use space to separate fields).  The automatic -c option can be\n"\
"   overridden by giving an explicit one later on the command line.\n";
    char FieldChar = '\n';

    while ((c = getopt(argc,argv,"adDhlnort:T:L:f:m:M:c:s:")) != -1)
	switch (c) {
	case 'a':
	    aOpt = 1;
	    break;
	case 'd':
	    Desc = 1;
	    break;
	case 'D':
	    ReadDirs = 1;
	    break;
	case 'o':
	    AllOpen = 1;
	    break;
	case 'n':
	    NameLen = 0;
	    break;
	case 'r':
	    ReadInput = 1;
	    break;
	case 's':
	    tagChar = *optarg;
	    gotTag = 1;
	    ReadInput = 1;
	    break;
	case 'h':
	    printf(Usage,AllFields);
	    printf(Help1,MAXTAG,TIME_FMT,AllFields,AllFields);
	    fputs(Help2,stdout);
	    exit(0);
	case 'L':
	    NameLen = 0;
	    FieldChar = ' ';
	case 'f':
	    Fields = optarg;
	    break;
	case 'T':
	    CheckType = 1;
	    switch (*optarg) {
	    case 's':
		CheckTypeRDev = S_INSEM;
		CheckTypeMode = S_IFNAM;
		break;
	    case 'm':
		CheckTypeRDev = S_INSHD;
		CheckTypeMode = S_IFNAM;
		break;
	    default:
		if (!FileType(*optarg, &CheckTypeMode, &CheckTypeRDev)) {
		    fprintf(stderr,
		    "stat: Unknown character passed with -%c: %c\n",c,*optarg);
		    Error = 1;
		}
	    }
	    break;
	case 'M':
	    switch (*optarg) {
	    case '-':
		AnyBits = 1;
		optarg++;
		break;
	    case '+':
		optarg++;
		AllBits = 1;
		break;
	    }
	    Error |= !(CheckModeMask = GetMode(optarg,(char) c));
	    break;
	case 'm':
	    if (!(ModeMask = GetMode(optarg,(char) c))) {
		Error = 1;
		break;
	    }
	    {
		mode_t MaskDup = ModeMask;
		
		/* Can't just use length of optarg for ModeZeros because
		   symbolic modes may be given */
		/* Count 1 bits for 'O' field spec and position of highest 1
		   for 'o' field spec */
		/* Start with count set to 2 so that when we divide by 3,
		   any fractional part will be rounded up. */
		for (ModeZeros = ModeXZeros = 2; MaskDup; MaskDup >>= 1) {
		    if (MaskDup & 1)
			ModeXZeros++;
		    ModeZeros++;
		}
		ModeZeros /= 3;
		ModeXZeros /= 3;
	    }
	    break;
	case 'c':
	    FieldChar = (char) (Check = StrToChar(optarg));
	    if (Check == -1) {
		fprintf(stderr,"stat: Invalid value given with -%c.\n",c);
		Error = 1;
	    }
	    break;
	case 't':
	    TimeFormat = optarg;
	    break;
	case 'l':
	    StatFunc = stat;
	    break;
	case '?':
	    Error = 1;
	}
    if (Error || !ReadInput && !AllOpen && optind >= argc) {
	fprintf(stderr,Usage,AllFields);
	exit(1);
    }
    if (aOpt) {
	char *AllFieldsEnd = AllFields + sizeof(AllFields) - 4;
	if (!ModeMask) {
	    *(AllFieldsEnd++) = 'O';
	    *(AllFieldsEnd++) = 'B';
	}
	if (tagChar)
	    *(AllFieldsEnd++) = 't';
	*AllFieldsEnd = '\0';
	Fields = AllFields;
    }
    if (InitFuncs(Funcs,Fields,gotTag))
	exit(1);
    for (; optind < argc; optind++)
	Error |= StatFile(argv[optind],0,Funcs,FieldChar,NameLen,Desc,ReadDirs);
    if (ReadInput) {
	while (Length = ngets(FileName,MAXNAMLEN+MAXTAG+1,stdin,
	gotTag && !tagChar)) {
	    if (gotTag)
		if (TagStart = lrindex(FileName,tagChar,Length)) {
		    Tag = TagStart + 1;
		    *TagStart = '\0';
		}
		else
		    Tag = "";
	    Error |= StatFile(FileName,0,Funcs,FieldChar,NameLen,Desc,ReadDirs);
	}
    }
    if (AllOpen)
	Error |= StatAllOpen(Funcs,FieldChar,NameLen);
    exit(Error);
}

/*
*   Return the last instance of character c in string s.
*   len is the length of string s.
*   This version of rindex will work with strings that contain null characters,
*   and can search for a null character.
*   Return value:
*   A pointer to the last instance of c in s.  If c does not occur in s,
*   returns a null pointer.
*/
char *lrindex(char *s, char c, unsigned len)
{
    char *p;

    if (!len)
	return (char *)0;
    for (p = s+len-1; p >= s; p--)
	if (*p == c)
	    return p;
    return (char *)0;
}

/*
*   ngets: read a string from an open file.
*   ngets reads from a file until it encounters a newline or the end of the
*   file, and returns the string read.  Trailing newlines are removed.
*   Strings that are empty (zero-length) after the trailing newline (if any)
*   is removed are discarded.
*   Input variables:
*   BufSize is the size of the buffer that the string is to be stored in.
*   Up to BufSize-2 characters are stored in the buffer.  If a newline is
*   not encountered before BufSize-2 characters have been read, they are
*   returned and the remainder are left for the next call.
*   f is the stdio file to read from.
*   If GetNulls is true, input is read with a series of fgetc() calls so that
*   null characters can be read.  If false, input is read with fgets(), which
*   cannot deal with nulls (since it does not return a read length) but is
*   more efficient.
*   Output variables:
*   The string is stored in Buf[].
*   Return value: The number of characters stored in buf, or zero at EOF
*   (since empty strings are discarded, a zero return always means EOF).
*/
unsigned ngets(char Buf[],unsigned BufSize,FILE *f,int GetNulls)
{
    unsigned Length;

    do {
	if (GetNulls) {
	    if (!(Length = cgets(Buf,BufSize-1,f)))
		return 0;
	}
	else {
	    if (!fgets(Buf,BufSize,f))
		return 0;
	    /* get rid of newline */
	    Length = strlen(Buf);
	}
	if (Buf[Length-1] == '\n')
	    Buf[--Length] = '\0';
    } while (!Length);
    return Length;
}

/*
*   ngets: read a string from an open file.
*   ngets reads from a file until it encounters a newline or the end of the
*   file, and returns the string read.  Trailing newlines are included.
*   The string is not null-terminated.
*   characters.
*   Input variables:
*   BufSize is the size of the buffer that the string is to be stored in.
*   Up to BufSize characters are stored in the buffer.  If a newline is
*   not encountered before BufSize characters have been read, they are
*   returned and the remainder are left for the next call.
*   f is the stdio file to read from.
*   Output variables:
*   The string is stored in Buf[].
*   Return value: The number of characters stored in buf, or zero at EOF.
*   Since the trailing newline is included in the length, and there will
*   always be a trailing newline except at EOF, a zero return always means
*   EOF (except that if BufSize is 0, 0 will also be returned).
*/
unsigned cgets(char *Buf,unsigned BufSize,FILE *f)
{
    char *p = Buf-1;
    char *endp = Buf + BufSize - 1;

    if (!BufSize)
	return 0;
    do {
	if ((*(++p) = fgetc(f)) == EOF) {
	    p--;
	    break;
	}
    } while (p < endp && *p != '\n');
    return p - Buf + 1;
}

StatAllOpen(char *(*Funcs[])(),char FieldChar,int NameLen)
{
    int fd;
    int NumFiles;
    int i;
    int Error = 0;

    NumFiles = (int) sysconf(_SC_OPEN_MAX);
    for (i = 0; i <= NumFiles; i++)
	Error |= StatFile(NULL,i,Funcs,FieldChar,NameLen,1,0);
}

/*
 *  Could avoid the need for this in some cases by using the info in the
 *  st_stat32 struct if that is what the stat() returned; it includes
 *  block size and fs type.
 */
struct statfs *DoStatfs()
{
    static struct statfs Buf;
    int Err;

    if (DidStatfs)
	return Buf.f_bsize ? &Buf : NULL;

    DidStatfs = 1;
    if (FileName ? statfs(FileName,&Buf,sizeof(Buf),0) :
    fstatfs(FileNum,&Buf,sizeof(Buf),0)) {
	fflush(stdout);
	/* Make sure we are at the start of a line */
	fputs("\nstat: ",stderr);
	perror(FileName);
	/* Make sure that if statfs is called again it will return NULL */
	Buf.f_bsize = 0;
	return NULL;
    }
    else
	return &Buf;
}
